/*
 Brad Garton's simple tutorial code for demonstrating neural network learning:
      - as created and demonstrated in Columbia University Computer Music Class of 2001.
      - reproduced with the Author's kind permission.
*/

#include <stdio.h>
#include "neural.h"

#define  ABS(x) ((x < 0) ? (-x) : (x))

struct output out;
struct hidden hid[NHIDDEN];
float in[NINPUT];

//----------------------------------------------------------------
int main()
{
	float tval,twt,target,err;
	int i,j,k;
	void srrand();
	float rrand();
	float brrand();
	float calcout();

/* connect 'em up and randomize the initial weights */
	for (i = 0; i < NHIDDEN; i++) {
		out.hidvals[i] = &(hid+i)->val;
		out.wts[i] = brrand();
		for (j = 0; j < NINPUT; j++) {
			hid[i].invals[j] = in+j;
			hid[i].wts[j] = brrand();
		}
	}

/* ok, we'll try a training */
	in[0] = 1.0;
	for (i = 1; i < NINPUT; i++) in[i] = 0.0;

	k = 1000;
	while (k--) {
		target = 1.0;
		tval = calcout();
		err = (ABS((target-tval)));

		/* now change the weights one by one and recalc */
		for (i = 0; i < NHIDDEN; i++) {
			for (j = 0; j < NINPUT; j++) {
				twt = hid[i].wts[j];
				/* try positive */
				hid[i].wts[j] += 0.001;
				tval = calcout();
				if (err <= (ABS((target-tval))) ) {
					/* try negative */
					hid[i].wts[j] -= 0.002;
					tval = calcout();
					if (err <= (ABS((target-tval))) ) {
						/* didn't work */
						hid[i].wts[j] = twt;
					}
				}
			}
		}
		tval = calcout();
		err = (ABS((target-tval)));

		/* now let's monkey with the hidden units */
		for (i = 0; i < NHIDDEN; i++) {
			twt = out.wts[i];
			/* try positive */
			out.wts[i] += 0.01;
			tval = calcout();
			if (err <= (ABS((target-tval))) ) {
				/* try negative */
				out.wts[i] -= 0.02;
				tval = calcout();
				if (err <= (ABS((target-tval))) ) {
					/* didn't work */
					out.wts[i] = twt;
				}
			}
		}
		tval = calcout();
		err = (ABS((target-tval)));
		printf("\ngeneration %d: ---------------------------- tval: %f; err: %f\n",k,tval,err);
		for (i = 0; i < NHIDDEN; i++) {
			printf("   h%d weight: %f\n",i,out.wts[i]);
			for (j = 0; j < NINPUT; j++) {
				printf("%f ",hid[i].wts[j]);
			}
			printf("\n");
		}
	}
}

//----------------------------------------------------------------
float calcout()
{
	int i,j;
	float rval = 0.0;

	for (i = 0; i < NHIDDEN; i++) {
		/* calculate the hidden values */
		hid[i].val = 0.0;
		for (j = 0; j < NINPUT; j++) {
			hid[i].val += (*(hid[i].invals[j]) * hid[i].wts[j]);
		}
	rval += (*(out.hidvals[i]) * out.wts[i]);
	}

	return(rval);
}


//----------------------------------------------------------------

long randx = 1;

void srrand(x)
unsigned x;
{
	randx = x;
}

//----------------------------------------------------------------
float rrand()
{
	int i = ((randx = randx*1103515245 + 12345)>>16) & 077777;
	return((float)i/16384.0 - 1.0);
}

//----------------------------------------------------------------
float brrand()
{
	int i = ((randx = randx*1103515245 + 12345)>>16) & 077777;
	return((float)i/32768.0);
}


