/*
 Brad Garton's simple tutorial code for demonstrating neural network learning:
      - as created and demonstrated in Columbia University Computer Music Class of 2001.
      - reproduced with the Author's kind permission.
*/
 
#include <stdio.h>
#include "neural.h"

#include <Xm/Xm.h>
#include <Xm/BulletinB.h>
#include <Xm/DrawingA.h>
#include <Xm/PushB.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/Label.h>


Widget theDraw,topLevel,bboard;
Widget initB,learnB,decideB;
Widget genText,genLabel;
Widget learnText,choiceLabel,choiceLabelN;

Display *theD;
Window   theW;

GC gc;
Colormap cmap;
XColor bground,positive[4],negative[4],plain,dots;
XColor smbground,lines,rectline,unused;

XtAppContext app;

int outxy[2];
int hidxy[NHIDDEN][2];
int inxy[NINPUT][2];
int rectxy[2];
int rectw,recth;
int linesxy[NINPUT][2];

int ngens;

#define  ABS(x) ((x < 0) ? (-x) : (x))

struct output out;
struct hidden hid[NHIDDEN];
float in[NINPUT];

//----------------------------------------------------------------
int main(argc,argv)
int argc;
char *argv[];
{
    XmString text;
	void GoInit();
	void GoLearn();
	void GoDecide();
	void GetGens();
	void srrand();
	float rrand();
	float brrand();
	int i;

	XtActionsRec actions;
	void movelines();
	String translations =
		"<Btn1Down>:	movelines()	ManagerGadgetArm()  \n\
		 <Btn1Up>:	movelines()	ManagerGadgetActivate() \n\
		 <Btn1Motion>:	movelines()	ManagerGadgetButtonMotion()";


	topLevel=XtVaAppInitialize(&app,"NeuralG",
		NULL,0,&argc,argv,NULL,NULL);

/* set up the main window */
	bboard = XtVaCreateManagedWidget("bboard",
		xmBulletinBoardWidgetClass, topLevel,
		XmNwidth, 800, XmNheight, 600,
		XmNmarginWidth, 0, XmNmarginHeight, 0,NULL);

	theD = XtDisplay(bboard);

	actions.string = "movelines";
	actions.proc = movelines;
	XtAppAddActions(app, &actions, 1);

/* set up the colors */
	cmap = DefaultColormapOfScreen(XtScreen(bboard));
	XAllocNamedColor (theD,cmap,"lightyellow3",&bground,&unused);
	XAllocNamedColor (theD,cmap,"dodgerblue1",&negative[3],&unused);
	XAllocNamedColor (theD,cmap,"dodgerblue2",&negative[2],&unused);
	XAllocNamedColor (theD,cmap,"dodgerblue3",&negative[1],&unused);
	XAllocNamedColor (theD,cmap,"dodgerblue4",&negative[0],&unused);
	XAllocNamedColor (theD,cmap,"red1",&positive[3],&unused);
	XAllocNamedColor (theD,cmap,"red2",&positive[2],&unused);
	XAllocNamedColor (theD,cmap,"red3",&positive[1],&unused);
	XAllocNamedColor (theD,cmap,"red4",&positive[0],&unused);
	XAllocNamedColor (theD,cmap,"palegreen4",&plain,&unused);
	XAllocNamedColor (theD,cmap,"purple4",&dots,&unused);
	XAllocNamedColor (theD,cmap,"chartreuse",&rectline,&unused);
	XAllocNamedColor (theD,cmap,"lightskyblue",&smbground,&unused);
	XAllocNamedColor (theD,cmap,"coral4",&lines,&unused);

	gc = XCreateGC(theD,RootWindowOfScreen(XtScreen(bboard)), 0, NULL);

/* set up the drawing area in the main window */
	theDraw = XtVaCreateManagedWidget("theDraw",
		xmDrawingAreaWidgetClass, bboard,
		XmNtranslations, XtParseTranslationTable(translations),
		XmNwidth, 800, XmNheight, 500,
		XmNbackground, bground.pixel, NULL);

/* init button */
	text = XmStringCreateSimple("initialize");
	initB = XtVaCreateManagedWidget("initbutton",
		xmPushButtonWidgetClass, bboard,
		XmNlabelString, text,
		XmNx, 80, XmNy, 510, NULL);
	XtAddCallback(initB, XmNactivateCallback, GoInit, NULL);

/* learn button */
	text = XmStringCreateSimple("  learn  ");
	learnB = XtVaCreateManagedWidget("learnbutton",
		xmPushButtonWidgetClass, bboard,
		XmNlabelString, text,
		XmNx, 80, XmNy, 545, NULL);
	XtAddCallback(learnB, XmNactivateCallback, GoLearn, NULL);

/* ngenerations text field */
	ngens = 1000;
	genText = XtVaCreateManagedWidget("gentext",
		xmTextFieldWidgetClass, bboard,
		XmNvalue, "1000",
		XmNwidth,70,XmNheight,35,
		XmNx, 355, XmNy, 535, NULL);
	XtAddCallback(genText, XmNactivateCallback, GetGens, NULL);

/* ngenerations label */
	text = XmStringCreateSimple("number of generations");
	genLabel = XtVaCreateManagedWidget("genlabel",
		xmLabelWidgetClass, bboard,
		XmNlabelString, text,
		XmNx, 320, XmNy, 575, NULL);

/* learning phase text display */
	learnText = XtVaCreateManagedWidget("learntext",
		xmTextFieldWidgetClass, bboard,
		XmNwidth, 420, XmNheight, 30,
		XmNeditable,0,XmNshadowThickness,0,
		XmNx, 230, XmNy, 505, NULL);
	
/* decide button */
	text = XmStringCreateSimple("decide");
	decideB = XtVaCreateManagedWidget("decidebutton",
		xmPushButtonWidgetClass, bboard,
		XmNlabelString, text,
		XmNx, 650, XmNy, 510, NULL);
	XtAddCallback(decideB, XmNactivateCallback, GoDecide, NULL);

/* choice label display 1 */
	text = XmStringCreateSimple("");
	choiceLabel = XtVaCreateManagedWidget("choicelabel",
		xmLabelWidgetClass, bboard,
		XmNlabelString, text,
		XmNx, 600, XmNy, 540, NULL);

/* choice label display 2 */
	text = XmStringCreateSimple("");
	choiceLabelN = XtVaCreateManagedWidget("choicelabelN",
		xmLabelWidgetClass, bboard,
		XmNlabelString, text,
		XmNx, 600, XmNy, 560, NULL);


/* set up the coordinates for the units and other stuff */
	outxy[0] = 400;
	outxy[1] = 200;
	for (i = 0; i < NHIDDEN; i++) {
		hidxy[i][0] = (i*400) + 200;
		hidxy[i][1] = 400;
	}
	for (i = 0; i < NINPUT; i++) {
		inxy[i][0] = (i*100)+50;
		inxy[i][1] = 480;
	}
	rectxy[0] = 500;
	rectxy[1] = 10;
	rectw = 200;
	recth = 200;
	for (i = 0; i < NINPUT; i++) {
		linesxy[i][0] = rectxy[0]+((i*25)+10);
		linesxy[i][1] = rectxy[1]+recth;
	}


	XtManageChild(bboard);
	XtManageChild(theDraw);
	XtRealizeWidget(topLevel);

	theW = XtWindow(theDraw);

	srrand(78);

	XtAppMainLoop(app);
}

//----------------------------------------------------------------
void GoInit(w, client, call)
Widget w;
XtPointer client, call;
{
        XmString text;
	int i,j;
	void srrand();
	float rrand();
	float brrand();

/* connect 'em up and randomize the initial weights */
	for (i = 0; i < NHIDDEN; i++) {
		out.hidvals[i] = &(hid+i)->val;
		out.wts[i] = brrand();
		for (j = 0; j < NINPUT; j++) {
			hid[i].invals[j] = in+j;
			hid[i].wts[j] = brrand();
		}
	}

/* draw the lines */
	XSetForeground(theD, gc, plain.pixel);
	for (i = 0; i < NHIDDEN; i++) {
		/* hidden to out */
		XDrawLine(theD,theW, gc, hidxy[i][0], hidxy[i][1],
			outxy[0],outxy[1]);
		for (j = 0; j < NINPUT; j++) {
			/* input to hidden */
			XDrawLine(theD,theW, gc, inxy[j][0], inxy[j][1],
				hidxy[i][0],hidxy[i][1]);
		}
	}

/* draw the units */
	XSetForeground(theD, gc, dots.pixel);
	XFillArc(theD,theW, gc, outxy[0]-2, outxy[1]-2,
		4, 4, 0, (360*64));
	for (i = 0; i < NHIDDEN; i++) {
		XFillArc(theD,theW, gc, hidxy[i][0]-2, hidxy[i][1]-2,
			4, 4, 0, (360*64));
		for (j = 0; j < NINPUT; j++) {
			XFillArc(theD,theW, gc, inxy[j][0]-2, inxy[j][1]-2,
				4, 4, 0, (360*64));
		}
	}

/* draw the input display */
	XSetForeground(theD, gc, smbground.pixel);
	XFillRectangle(theD,theW, gc, rectxy[0], rectxy[1], rectw, recth);
	XSetForeground(theD, gc, rectline.pixel);
	XDrawRectangle(theD,theW, gc, rectxy[0], rectxy[1], rectw, recth);

/* blitz out some stuff */
	text=XmStringCreateSimple("");
	XtVaSetValues(choiceLabel,XmNlabelString,text,NULL);
	text=XmStringCreateSimple("");
	XtVaSetValues(choiceLabelN,XmNlabelString,text,NULL);
}

//----------------------------------------------------------------

void GoLearn(w, client, call)
Widget w;
XtPointer client, call;
{
	float tval,twt,target,err;
	float outbigneg,outbigpos,inbigneg,inbigpos;
	int i,j,k;
	int flipper,ival;
	void srrand();
	float rrand();
	float brrand();
	float calcout();
	char cvals[50];
	XmString text;


	outbigneg = outbigpos = inbigneg = inbigpos = 0.0;
	flipper = 0;
	for (k = 0; k < ngens; k++) {
		/* generate a test set, flip from low to high */
		if (flipper++ < 50) { /* low set */
			target = 1.0;
			for (i = 0; i < NINPUT/2; i++)
				in[i] = (brrand() * 0.3) + 0.7;
			for (i = NINPUT/2; i < NINPUT; i++)
				in[i] = brrand() * 0.3;
		}
		else {
			target = -1.0;
			for (i = 0; i < NINPUT/2; i++)
				in[i] = brrand() * 0.3;
			for (i = NINPUT/2; i < NINPUT; i++)
				in[i] = (brrand() * 0.3) + 0.7;
			if (flipper > 100) flipper = 0;
		}

		if ((flipper == 25) || (flipper == 75)) {
			sprintf(cvals,"gen #: %d  target: %f  value: %f",k,target,tval);
			XmTextSetString(learnText,cvals);
			XFlush(theD);
		}

		/* plot the input in the little window */
		for (i = 0; i < NINPUT; i++) {
			XSetForeground(theD, gc, smbground.pixel);
			XDrawLine(theD,theW, gc, linesxy[i][0], linesxy[i][1],
				linesxy[i][0],linesxy[i][1]-recth);
			XSetForeground(theD, gc, lines.pixel);
			ival = in[i] * (float)recth;
			XDrawLine(theD,theW, gc, linesxy[i][0], linesxy[i][1],
				linesxy[i][0],linesxy[i][1]-ival);
		}


		tval = calcout();
		err = (ABS((target-tval)));

		/* now change the weights one by one and recalc */
		for (i = 0; i < NHIDDEN; i++) {
			for (j = 0; j < NINPUT; j++) {
				twt = hid[i].wts[j];
				/* try positive */
				hid[i].wts[j] += 0.001;
				tval = calcout();
				if (err <= (ABS((target-tval))) ) {
					/* try negative */
					hid[i].wts[j] -= 0.002;
					tval = calcout();
					if (err <= (ABS((target-tval))) ) {
						/* didn't work */
						hid[i].wts[j] = twt;
					}
				}
				/* check for biggest neg or pos (graphing) */
				if (hid[i].wts[j] < inbigneg) 
					inbigneg = hid[i].wts[j];
				if (hid[i].wts[j] > inbigpos)
					inbigpos = hid[i].wts[j];
			}
		}
		tval = calcout();
		err = (ABS((target-tval)));

		/* now let's monkey with the hidden units */
		for (i = 0; i < NHIDDEN; i++) {
			twt = out.wts[i];
			/* try positive */
			out.wts[i] += 0.005;
			tval = calcout();
			if (err <= (ABS((target-tval))) ) {
				/* try negative */
				out.wts[i] -= 0.01;
				tval = calcout();
				if (err <= (ABS((target-tval))) ) {
					/* didn't work */
					out.wts[i] = twt;
				}
			}
			if (out.wts[i] < outbigneg)
				outbigneg = out.wts[i];
			if (out.wts[i] > outbigpos)
				outbigpos = out.wts[i];
		}
		tval = calcout();
		err = (ABS((target-tval)));
/*
		printf("\ngeneration %d: -------------- targ: %f; tval: %f; err: %f\n",k,target,tval,err);
		for (i = 0; i < NHIDDEN; i++) {
			printf("   h%d weight: %f\n",i,out.wts[i]);
			for (j = 0; j < NINPUT; j++) {
				printf("%f ",hid[i].wts[j]);
			}
			printf("\n");
		}
*/

		/* draw the lines */
		for (i = 0; i < NHIDDEN; i++) {
			/* hidden to out */
			if (out.wts[i] < 0) {
				if (out.wts[i] > (0.25*outbigneg))
					XSetForeground(theD,gc,negative[0].pixel);
				else if (out.wts[i] > (0.5*outbigneg))
					XSetForeground(theD,gc,negative[1].pixel);
				else if (out.wts[i] > (0.75*outbigneg))
					XSetForeground(theD,gc,negative[2].pixel);
				else XSetForeground(theD,gc,negative[3].pixel);
			}
			if (out.wts[i] > 0) {
				if (out.wts[i] > (0.75*outbigpos))
					XSetForeground(theD,gc,positive[3].pixel);
				else if (out.wts[i] > (0.5*outbigpos))
					XSetForeground(theD,gc,positive[2].pixel);
				else if (out.wts[i] > (0.25*outbigpos))
					XSetForeground(theD,gc,positive[1].pixel);
				else XSetForeground(theD,gc,positive[0].pixel);
			}
			XDrawLine(theD,theW, gc, hidxy[i][0], hidxy[i][1],
				outxy[0],outxy[1]);
			for (j = 0; j < NINPUT; j++) {
				/* input to hidden */
				if (hid[i].wts[j] < 0) {
					if (hid[i].wts[j] > (0.25*inbigneg))
						XSetForeground(theD,gc,negative[0].pixel);
					else if (hid[i].wts[j] > (0.5*inbigneg))
						XSetForeground(theD,gc,negative[1].pixel);
					else if (hid[i].wts[j] > (0.75*inbigneg))
						XSetForeground(theD,gc,negative[2].pixel);
					else XSetForeground(theD,gc,negative[3].pixel);
				}
				if (hid[i].wts[j] > 0) {
					if (hid[i].wts[j] > (0.75*inbigpos))
						XSetForeground(theD,gc,positive[3].pixel);
					else if (hid[i].wts[j] > (0.5*inbigpos))
						XSetForeground(theD,gc,positive[2].pixel);
					else if (hid[i].wts[j] > (0.25*inbigpos))
						XSetForeground(theD,gc,positive[1].pixel);
					else XSetForeground(theD,gc,positive[0].pixel);
				}
				XDrawLine(theD,theW, gc, inxy[j][0], inxy[j][1],
					hidxy[i][0],hidxy[i][1]);
			}
		}

		/* draw the units */
		XSetForeground(theD, gc, dots.pixel);
		XFillArc(theD,theW, gc, outxy[0]-2, outxy[1]-2,
			4, 4, 0, (360*64));
		for (i = 0; i < NHIDDEN; i++) {
			XFillArc(theD,theW, gc, hidxy[i][0]-2, hidxy[i][1]-2,
				4, 4, 0, (360*64));
			for (j = 0; j < NINPUT; j++) {
				XFillArc(theD,theW, gc, inxy[j][0]-2, inxy[j][1]-2,
					4, 4, 0, (360*64));
			}
		}
	}

	sprintf(cvals,"gen #: %d  target: %f  value: %f",k,target,tval);
	XmTextSetString(learnText,cvals);
	XFlush(theD);
}
//----------------------------------------------------------------
void GoDecide(w, client, call)
Widget w;
XtPointer client, call;
{
	float choice;
	float calcout();
	XmString text;
	char cvals[20];

	choice = calcout();
	if (choice > 0.0) {
		text=XmStringCreateSimple("This is a LOW sound");
		XtVaSetValues(choiceLabel,XmNlabelString,text,NULL);
		sprintf(cvals,"value: %f",choice);
		text=XmStringCreateSimple(cvals);
		XtVaSetValues(choiceLabelN,XmNlabelString,text,NULL);
	}
	else {
		text=XmStringCreateSimple("This is a HIGH sound");
		XtVaSetValues(choiceLabel,XmNlabelString,text,NULL);
		sprintf(cvals,"value: %f",choice);
		text=XmStringCreateSimple(cvals);
		XtVaSetValues(choiceLabelN,XmNlabelString,text,NULL);
	}
	XFlush(theD);
}
//----------------------------------------------------------------
void GetGens(w, client, call)
Widget w;
XtPointer client, call;
{
	char *text;
	extern int atoi();
	char tval[5];

	text=XmTextGetString(genText);
	ngens = atoi(text);
}

//----------------------------------------------------------------
float calcout()
{
	int i,j;
	float rval = 0.0;

	for (i = 0; i < NHIDDEN; i++) {
		/* calculate the hidden values */
		hid[i].val = 0.0;
		for (j = 0; j < NINPUT; j++) {
			hid[i].val += (*(hid[i].invals[j]) * hid[i].wts[j]);
		}
	rval += (*(out.hidvals[i]) * out.wts[i]);
	}

	return(rval);
}
//----------------------------------------------------------------
void movelines(w, event, args, nargs)
Widget w;
XEvent *event;
String *args;
int *nargs;
{
	Position x,y;
	int i;
	static int selected = -1;

	if (event->xany.type == ButtonPress) {
		x = event->xbutton.x;
		y = event->xbutton.y;

		if ( (x > rectxy[0]) && (x < (rectxy[0]+rectw)) &&
			(y > rectxy[1]) && (y < (rectxy[1]+recth)) ) {

			i = 0;
			while (x > (linesxy[i][0]+5)) i++;
			selected = i;
			XSetForeground(theD, gc, smbground.pixel);
			XDrawLine(theD,theW, gc, linesxy[i][0], linesxy[i][1],
				linesxy[i][0],linesxy[i][1]-recth);
			XSetForeground(theD, gc, lines.pixel);
			XDrawLine(theD,theW, gc, linesxy[i][0], linesxy[i][1],
				linesxy[i][0],y);
		}
	}

	if ((event->xany.type == ButtonRelease) && (selected > -1)) {
		x = event->xbutton.x;
		y = event->xbutton.y;
		if (y >= linesxy[selected][1]) y = linesxy[selected][1];
		if (y <= rectxy[1]) y = rectxy[1];
		XSetForeground(theD, gc, smbground.pixel);
		XDrawLine(theD,theW, gc, 
			linesxy[selected][0], linesxy[selected][1],
			linesxy[selected][0], linesxy[selected][1]-recth);
		XSetForeground(theD, gc, lines.pixel);
		XDrawLine(theD,theW, gc, 
			linesxy[selected][0], linesxy[selected][1],
			linesxy[selected][0], y);

		XSetForeground(theD, gc, rectline.pixel);
		XDrawRectangle(theD,theW,gc,rectxy[0],rectxy[1],rectw,recth);

		in[selected] = ((float)(linesxy[selected][1]-y))/(float)recth;

		selected = -1;
	}

	if ((event->xany.type == MotionNotify) && (selected > -1)) {
		x = event->xbutton.x;
		y = event->xbutton.y;
		if (y >= linesxy[selected][1]) y = linesxy[selected][1];
		if (y <= rectxy[1]) y = rectxy[1];
		XSetForeground(theD, gc, smbground.pixel);
		XDrawLine(theD,theW, gc, 
			linesxy[selected][0], linesxy[selected][1],
			linesxy[selected][0], linesxy[selected][1]-recth);
		XSetForeground(theD, gc, lines.pixel);
		XDrawLine(theD,theW, gc, 
			linesxy[selected][0], linesxy[selected][1],
			linesxy[selected][0], y);
	}
}


//----------------------------------------------------------------
long randx = 1;
//----------------------------------------------------------------
void srrand(x)
unsigned x;
{
	randx = x;
}
//----------------------------------------------------------------
float rrand()
{
	int i = ((randx = randx*1103515245 + 12345)>>16) & 077777;
	return((float)i/16384.0 - 1.0);
}
//----------------------------------------------------------------
float brrand()
{
	int i = ((randx = randx*1103515245 + 12345)>>16) & 077777;
	return((float)i/32768.0);
}
//----------------------------------------------------------------

